# Copyright (c) 2012-2014 Erik Johansson <erik@ejohansson.se>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

from ctypes import c_bool, c_char_p, c_int, c_ubyte, c_void_p
from ctypes import byref, cast, create_string_buffer, POINTER, sizeof
import platform
import threading

import tellcore.constants as const


if platform.system() == 'Windows':
    from ctypes import WINFUNCTYPE as FUNCTYPE, windll as DllLoader
    LIBRARY_NAME = 'TelldusCore.dll'
else:
    from ctypes import CFUNCTYPE as FUNCTYPE, cdll as DllLoader
    if platform.system() == 'Darwin':
        from ctypes.util import find_library
        LIBRARY_NAME = find_library('TelldusCore') or \
            '/Library/Frameworks/TelldusCore.framework/TelldusCore'
    else:
        LIBRARY_NAME = 'libtelldus-core.so.2'

DEVICE_EVENT_FUNC = FUNCTYPE(
    None, c_int, c_int, c_char_p, c_int, c_void_p)
DEVICE_CHANGE_EVENT_FUNC = FUNCTYPE(
    None, c_int, c_int, c_int, c_int, c_void_p)
RAW_DEVICE_EVENT_FUNC = FUNCTYPE(
    None, c_char_p, c_int, c_int, c_void_p)
SENSOR_EVENT_FUNC = FUNCTYPE(
    None, c_char_p, c_char_p, c_int, c_int, c_char_p, c_int, c_int, c_void_p)
CONTROLLER_EVENT_FUNC = FUNCTYPE(
    None, c_int, c_int, c_int, c_char_p, c_int, c_void_p)


class TelldusError(Exception):
    """Error returned from Telldus Core API.

    Automatically raised when a function in the C API returns an error.

    Attributes:
        error: The error code constant (one of TELLSTICK_ERROR_* from
        :mod:`tellcore.constants`).
    """

    def __init__(self, error, lib=None):
        super(TelldusError, self).__init__()
        self.error = error
        self.lib = lib or Library()

    def __str__(self):
        """Return the human readable error string."""
        msg = self.lib.tdGetErrorString(self.error)
        return "%s (%d)" % (msg, self.error)


class BaseCallbackDispatcher(object):
    """Base callback dispatcher class.

    Inherit from this class and override the :func:`on_callback` method to
    change how callbacks are dispatched.
    """

    def on_callback(self, callback, *args):
        """Called from the callback thread when an event is received.

        :param callable callback: The callback function to call.
        :param args: The arguments to pass to the callback.
        """
        raise NotImplementedError


class DirectCallbackDispatcher(BaseCallbackDispatcher):
    """Dispatches callbacks directly.

    Since the callback is dispatched directly, the callback is called in the
    callback thread.
    """

    def on_callback(self, callback, *args):
        callback(*args)


class Library(object):
    """Wrapper around the Telldus Core C API.

    With the exception of tdInit, tdClose and tdReleaseString, all functions in
    the C API (see `Telldus Core documentation
    <http://developer.telldus.com/doxygen/group__core.html>`_) can be
    called. The parameters are the same as in the C API documentation. The
    return value are mostly the same as for the C API, except for functions
    with multiple out parameters.

    In addition, this class:
       * automatically frees memory for strings returned from the C API,
       * converts errors returned from functions into
         (:class:`TelldusError`) exceptions,
       * transparently converts between Python strings and C style strings.
    """

    STRING_ENCODING = 'utf-8'

    class c_string_p(c_char_p):
        def __init__(self, param):
            c_char_p.__init__(self, param.encode(Library.STRING_ENCODING))

        @classmethod
        def from_param(cls, param):
            if type(param) is str:
                return cls(param)
            try:
                if type(param) is unicode:
                    return cls(param)
            except NameError:
                pass  # The unicode type does not exist in python 3
            return c_char_p.from_param(param)

    # Must be a separate class (i.e. not part of Library), to avoid circular
    # references when saving the wrapper callback function in a class with a
    # destructor, as the destructor is not called in that case.
    class CallbackWrapper(object):
        def __init__(self, dispatcher):
            self._callbacks = {}
            self._lock = threading.Lock()
            self._dispatcher = dispatcher

        def get_callback_ids(self):
            with self._lock:
                return list(self._callbacks.keys())

        def register_callback(self, registrator, functype, callback):
            wrapper = functype(self._callback)
            with self._lock:
                cid = registrator(wrapper, None)
                self._callbacks[cid] = (wrapper, callback)
                return cid

        def unregister_callback(self, cid):
            with self._lock:
                del self._callbacks[cid]

        def _callback(self, *in_args):
            args = []
            # Convert all char* parameters (i.e. bytes) to proper python
            # strings
            for arg in in_args:
                if type(arg) is bytes:
                    args.append(arg.decode(Library.STRING_ENCODING))
                else:
                    args.append(arg)

            # Get the real callback and the dispatcher
            with self._lock:
                try:
                    # args[-2] is callback id
                    (wrapper, callback) = self._callbacks[args[-2]]
                except KeyError:
                    return
                dispatcher = self._dispatcher

            # Dispatch the callback, dropping the last parameter which is the
            # context and always None.
            try:
                dispatcher.on_callback(callback, *args[:-1])
            except:
                pass

    _lib = None
    _refcount = 0

    _functions = {
        'tdInit': [None, []],
        'tdClose': [None, []],
        'tdReleaseString': [None, [c_void_p]],
        'tdGetErrorString': [c_char_p, [c_int]],

        'tdRegisterDeviceEvent':
        [c_int, [DEVICE_EVENT_FUNC, c_void_p]],
        'tdRegisterDeviceChangeEvent':
        [c_int, [DEVICE_CHANGE_EVENT_FUNC, c_void_p]],
        'tdRegisterRawDeviceEvent':
        [c_int, [RAW_DEVICE_EVENT_FUNC, c_void_p]],
        'tdRegisterSensorEvent':
        [c_int, [SENSOR_EVENT_FUNC, c_void_p]],
        'tdRegisterControllerEvent':
        [c_int, [CONTROLLER_EVENT_FUNC, c_void_p]],
        'tdUnregisterCallback': [c_int, [c_int]],

        'tdTurnOn': [c_int, [c_int]],
        'tdTurnOff': [c_int, [c_int]],
        'tdBell': [c_int, [c_int]],
        'tdDim': [c_int, [c_int, c_ubyte]],
        'tdExecute': [c_int, [c_int]],
        'tdUp': [c_int, [c_int]],
        'tdDown': [c_int, [c_int]],
        'tdStop': [c_int, [c_int]],
        'tdLearn': [c_int, [c_int]],
        'tdMethods': [c_int, [c_int, c_int]],
        'tdLastSentCommand': [c_int, [c_int, c_int]],
        'tdLastSentValue': [c_char_p, [c_int]],

        'tdGetNumberOfDevices': [c_int, []],
        'tdGetDeviceId': [c_int, [c_int]],
        'tdGetDeviceType': [c_int, [c_int]],

        'tdGetName': [c_char_p, [c_int]],
        'tdSetName': [c_bool, [c_int, c_string_p]],
        'tdGetProtocol': [c_char_p, [c_int]],
        'tdSetProtocol': [c_bool, [c_int, c_string_p]],
        'tdGetModel': [c_char_p, [c_int]],
        'tdSetModel': [c_bool, [c_int, c_string_p]],

        'tdGetDeviceParameter': [c_char_p, [c_int, c_string_p, c_string_p]],
        'tdSetDeviceParameter': [c_bool, [c_int, c_string_p, c_string_p]],

        'tdAddDevice': [c_int, []],
        'tdRemoveDevice': [c_bool, [c_int]],

        'tdSendRawCommand': [c_int, [c_string_p, c_int]],

        'tdConnectTellStickController': [None, [c_int, c_int, c_string_p]],
        'tdDisconnectTellStickController': [None, [c_int, c_int, c_string_p]],

        'tdSensor': [c_int, [c_char_p, c_int, c_char_p, c_int,
                             POINTER(c_int), POINTER(c_int)]],
        'tdSensorValue': [c_int, [c_string_p, c_string_p, c_int, c_int,
                                  c_char_p, c_int, POINTER(c_int)]],

        'tdController': [c_int, [POINTER(c_int), POINTER(c_int),
                                 c_char_p, c_int, POINTER(c_int)]],
        'tdControllerValue': [c_int, [c_int, c_string_p, c_char_p, c_int]],
        'tdSetControllerValue': [c_int, [c_int, c_string_p, c_string_p]],
        'tdRemoveController': [c_int, [c_int]],
    }

    def _to_str(self, char_p):
        return char_p.value.decode(Library.STRING_ENCODING)

    def _setup_functions(self, lib):
        def check_int_result(result, func, args):
            if result < 0:
                raise TelldusError(result)
            return result

        def check_bool_result(result, func, args):
            if not result:
                raise TelldusError(const.TELLSTICK_ERROR_DEVICE_NOT_FOUND)
            return result

        def free_string(result, func, args):
            string = cast(result, c_char_p).value
            if string is not None:
                lib.tdReleaseString(result)
                string = string.decode(Library.STRING_ENCODING)
            return string

        for name, signature in Library._functions.items():
            try:
                func = getattr(lib, name)
                func.restype = signature[0]
                func.argtypes = signature[1]

                if func.restype == c_int:
                    func.errcheck = check_int_result
                elif func.restype == c_bool:
                    func.errcheck = check_bool_result
                elif func.restype == c_char_p:
                    func.restype = c_void_p
                    func.errcheck = free_string
            except AttributeError:
                # Older version of the lib don't have all the functions
                pass

    def __init__(self, name=None, callback_dispatcher=None):
        """Load and initialize the Telldus core library.

        The underlaying library is only initialized the first time this object
        is created. Subsequent instances uses the same underlaying library
        instance.

        :param str name: If None than the platform specific name of the
            Telldus library is used, but it can be e.g. an absolute path.

        :param callback_dispatcher: If callbacks are to be used, this parameter
            must refer to an instance of a class inheriting from
            :class:`BaseCallbackDispatcher`.

        """
        super(Library, self).__init__()

        if not Library._lib:
            assert Library._refcount == 0
            if name is None:
                name = LIBRARY_NAME

            lib = DllLoader.LoadLibrary(name)
            self._setup_functions(lib)
            lib.tdInit()
            Library._lib = lib

        Library._refcount += 1
        if callback_dispatcher is not None:
            self._callback_wrapper = Library.CallbackWrapper(
                callback_dispatcher)
        else:
            self._callback_wrapper = None

    def __del__(self):
        """Close and unload the Telldus core library.

        Any callback set up is removed.

        The underlaying library is only closed and unloaded if this is the last
        instance sharing the same underlaying library instance.
        """
        # Happens if the LoadLibrary call fails
        if not Library._lib:
            assert Library._refcount == 0
            return

        assert Library._refcount >= 1
        Library._refcount -= 1

        if self._callback_wrapper is not None:
            for cid in self._callback_wrapper.get_callback_ids():
                try:
                    self.tdUnregisterCallback(cid)
                except:
                    pass

        if Library._refcount != 0:
            return

        # telldus-core before v2.1.2 (where tdController was added) does not
        # handle re-initialization after tdClose has been called (see Telldus
        # ticket 188).
        if hasattr(Library._lib, "tdController"):
            Library._lib.tdClose()
        Library._lib = None

    def __getattr__(self, name):
        if name == 'callback_dispatcher':
            return self._callback_wrapper._dispatcher
        if name in Library._functions:
            return getattr(self._lib, name)
        raise AttributeError(name)

    def tdInit(self):
        raise NotImplementedError('should not be called explicitly')

    def tdClose(self):
        raise NotImplementedError('should not be called explicitly')

    def tdReleaseString(self, string):
        raise NotImplementedError('should not be called explicitly')

    def tdRegisterDeviceEvent(self, callback):
        assert(self._callback_wrapper is not None)
        return self._callback_wrapper.register_callback(
            self._lib.tdRegisterDeviceEvent, DEVICE_EVENT_FUNC, callback)

    def tdRegisterDeviceChangeEvent(self, callback):
        assert(self._callback_wrapper is not None)
        return self._callback_wrapper.register_callback(
            self._lib.tdRegisterDeviceChangeEvent, DEVICE_CHANGE_EVENT_FUNC,
            callback)

    def tdRegisterRawDeviceEvent(self, callback):
        assert(self._callback_wrapper is not None)
        return self._callback_wrapper.register_callback(
            self._lib.tdRegisterRawDeviceEvent, RAW_DEVICE_EVENT_FUNC,
            callback)

    def tdRegisterSensorEvent(self, callback):
        assert(self._callback_wrapper is not None)
        return self._callback_wrapper.register_callback(
            self._lib.tdRegisterSensorEvent, SENSOR_EVENT_FUNC, callback)

    def tdRegisterControllerEvent(self, callback):
        assert(self._callback_wrapper is not None)
        return self._callback_wrapper.register_callback(
            self._lib.tdRegisterControllerEvent, CONTROLLER_EVENT_FUNC,
            callback)

    def tdUnregisterCallback(self, cid):
        assert(self._callback_wrapper is not None)
        self._callback_wrapper.unregister_callback(cid)
        self._lib.tdUnregisterCallback(cid)

    def tdSensor(self):
        """Get the next sensor while iterating.

        :return: a dict with the keys: protocol, model, id, datatypes.
        """
        protocol = create_string_buffer(20)
        model = create_string_buffer(20)
        sid = c_int()
        datatypes = c_int()

        self._lib.tdSensor(protocol, sizeof(protocol), model, sizeof(model),
                           byref(sid), byref(datatypes))
        return {'protocol': self._to_str(protocol),
                'model': self._to_str(model),
                'id': sid.value, 'datatypes': datatypes.value}

    def tdSensorValue(self, protocol, model, sid, datatype):
        """Get the sensor value for a given sensor.

        :return: a dict with the keys: value, timestamp.
        """
        value = create_string_buffer(20)
        timestamp = c_int()

        self._lib.tdSensorValue(protocol, model, sid, datatype,
                                value, sizeof(value), byref(timestamp))
        return {'value': self._to_str(value), 'timestamp': timestamp.value}

    def tdController(self):
        """Get the next controller while iterating.

        :return: a dict with the keys: id, type, name, available.
        """
        cid = c_int()
        ctype = c_int()
        name = create_string_buffer(255)
        available = c_int()

        self._lib.tdController(byref(cid), byref(ctype), name, sizeof(name),
                               byref(available))
        return {'id': cid.value, 'type': ctype.value,
                'name': self._to_str(name), 'available': available.value}

    def tdControllerValue(self, cid, name):
        value = create_string_buffer(255)

        self._lib.tdControllerValue(cid, name, value, sizeof(value))
        return self._to_str(value)



# Copyright (c) 2012-2014 Erik Johansson <erik@ejohansson.se>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

"""
Contains all constants used in the API.

All the TELLSTICK_* defines documented in the `Telldus Core documentation
<http://developer.telldus.com/doxygen/group__core.html>`_ are available here.

The recommended usage is to do::
    import tellcore.constants as const

The constants will then be available as const.TELLSTICK_TURNON, etc. See below
for the full list of all available constants.
"""

# Tellstick constants
# Device methods
TELLSTICK_TURNON = 1
TELLSTICK_TURNOFF = 2
TELLSTICK_BELL = 4
TELLSTICK_TOGGLE = 8
TELLSTICK_DIM = 16
TELLSTICK_LEARN = 32
TELLSTICK_EXECUTE = 64
TELLSTICK_UP = 128
TELLSTICK_DOWN = 256
TELLSTICK_STOP = 512

# Sensor value types
TELLSTICK_TEMPERATURE = 1
TELLSTICK_HUMIDITY = 2
TELLSTICK_RAINRATE = 4
TELLSTICK_RAINTOTAL = 8
TELLSTICK_WINDDIRECTION = 16
TELLSTICK_WINDAVERAGE = 32
TELLSTICK_WINDGUST = 64

# Error codes
TELLSTICK_SUCCESS = 0
TELLSTICK_ERROR_NOT_FOUND = -1
TELLSTICK_ERROR_PERMISSION_DENIED = -2
TELLSTICK_ERROR_DEVICE_NOT_FOUND = -3
TELLSTICK_ERROR_METHOD_NOT_SUPPORTED = -4
TELLSTICK_ERROR_COMMUNICATION = -5
TELLSTICK_ERROR_CONNECTING_SERVICE = -6
TELLSTICK_ERROR_UNKNOWN_RESPONSE = -7
TELLSTICK_ERROR_SYNTAX = -8
TELLSTICK_ERROR_BROKEN_PIPE = -9
TELLSTICK_ERROR_COMMUNICATING_SERVICE = -10
TELLSTICK_ERROR_CONFIG_SYNTAX = -11
TELLSTICK_ERROR_UNKNOWN = -99

# Device types
TELLSTICK_TYPE_DEVICE = 1
TELLSTICK_TYPE_GROUP = 2
TELLSTICK_TYPE_SCENE = 3

# Controller types
TELLSTICK_CONTROLLER_TELLSTICK = 1
TELLSTICK_CONTROLLER_TELLSTICK_DUO = 2
TELLSTICK_CONTROLLER_TELLSTICK_NET = 3

# Device changes
TELLSTICK_DEVICE_ADDED = 1
TELLSTICK_DEVICE_CHANGED = 2
TELLSTICK_DEVICE_REMOVED = 3
TELLSTICK_DEVICE_STATE_CHANGED = 4

# Change types
TELLSTICK_CHANGE_NAME = 1
TELLSTICK_CHANGE_PROTOCOL = 2
TELLSTICK_CHANGE_MODEL = 3
TELLSTICK_CHANGE_METHOD = 4
TELLSTICK_CHANGE_AVAILABLE = 5
TELLSTICK_CHANGE_FIRMWARE = 6



# Copyright (c) 2012-2014 Erik Johansson <erik@ejohansson.se>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

try:
    import queue
except ImportError:
    # Fall back on old (python 2) variant
    import Queue as queue

import tellcore.constants as const
from tellcore.library import Library, TelldusError, BaseCallbackDispatcher


class QueuedCallbackDispatcher(BaseCallbackDispatcher):
    """The default callback dispatcher used by :class:`TelldusCore`.

    Queues callbacks that arrive from Telldus Core. Then calls them in the main
    thread (or more precise: the thread calling :func:`process_callback`)
    instead of the callback thread used by Telldus Core. This way the
    application using :class:`TelldusCore` don't have to do any thread
    synchronization. Only make sure :func:`process_pending_callbacks` is called
    regularly.
    """

    def __init__(self):
        super(QueuedCallbackDispatcher, self).__init__()
        self._queue = queue.Queue()

    def on_callback(self, callback, *args):
        self._queue.put((callback, args))

    def process_callback(self, block=True):
        """Dispatch a single callback in the current thread.

        :param boolean block: If True, blocks waiting for a callback to come.
        :return: True if a callback was processed; otherwise False.
        """
        try:
            (callback, args) = self._queue.get(block=block)
            try:
                callback(*args)
            finally:
                self._queue.task_done()
        except queue.Empty:
            return False
        return True

    def process_pending_callbacks(self):
        """Dispatch all pending callbacks in the current thread."""
        while self.process_callback(block=False):
            pass


class AsyncioCallbackDispatcher(BaseCallbackDispatcher):
    """Dispatcher for use with the event loop available in Python 3.4+.

    Callbacks will be dispatched on the thread running the event loop. The loop
    argument should be a BaseEventLoop instance, e.g. the one returned from
    asyncio.get_event_loop().
    """
    def __init__(self, loop):
        super(AsyncioCallbackDispatcher, self).__init__()
        self._loop = loop

    def on_callback(self, callback, *args):
        self._loop.call_soon_threadsafe(callback, *args)


class TelldusCore(object):
    """The main class for tellcore-py.

    Has methods for adding devices and for enumerating controllers, devices and
    sensors. Also handles callbacks; both registration and making sure the
    callbacks are processed in the main thread instead of the callback thread.
    """

    def __init__(self, library_path=None, callback_dispatcher=None):
        """Create a new TelldusCore instance.

        Only one instance should be used per program.

        :param str library_path: Passed to the :class:`.library.Library`
            constructor.

        :param str callback_dispatcher: An instance implementing the
            :class:`.library.BaseCallbackDispatcher` interface (
            e.g. :class:`QueuedCallbackDispatcher` or
            :class:`AsyncioCallbackDispatcher`) A callback dispatcher must be
            provided if callbacks are to be used.

        """
        super(TelldusCore, self).__init__()
        self.lib = Library(library_path, callback_dispatcher)

    def __getattr__(self, name):
        if name == 'callback_dispatcher':
            return self.lib.callback_dispatcher
        raise AttributeError(name)

    def register_device_event(self, callback):
        """Register a new device event callback handler.

        See :ref:`event-example` for more information.

        :return: the callback id
        """
        return self.lib.tdRegisterDeviceEvent(callback)

    def register_device_change_event(self, callback):
        """Register a new device change event callback handler.

        See :ref:`event-example` for more information.

        :return: the callback id
        """
        return self.lib.tdRegisterDeviceChangeEvent(callback)

    def register_raw_device_event(self, callback):
        """Register a new raw device event callback handler.

        See :ref:`event-example` for more information.

        :return: the callback id
        """
        return self.lib.tdRegisterRawDeviceEvent(callback)

    def register_sensor_event(self, callback):
        """Register a new sensor event callback handler.

        See :ref:`event-example` for more information.

        :return: the callback id
        """
        return self.lib.tdRegisterSensorEvent(callback)

    def register_controller_event(self, callback):
        """Register a new controller event callback handler.

        See :ref:`event-example` for more information.

        :return: the callback id
        """
        return self.lib.tdRegisterControllerEvent(callback)

    def unregister_callback(self, cid):
        """Unregister a callback handler.

        :param int id: the callback id as returned from one of the
            register_*_event methods.
        """
        self.lib.tdUnregisterCallback(cid)

    def devices(self):
        """Return all known devices.

        :return: list of :class:`Device` or :class:`DeviceGroup` instances.
        """
        devices = []
        count = self.lib.tdGetNumberOfDevices()
        for i in range(count):
            device = DeviceFactory(self.lib.tdGetDeviceId(i), lib=self.lib)
            devices.append(device)
        return devices

    def sensors(self):
        """Return all known sensors.

        :return: list of :class:`Sensor` instances.
        """
        sensors = []
        try:
            while True:
                sensor = self.lib.tdSensor()
                sensors.append(Sensor(lib=self.lib, **sensor))
        except TelldusError as e:
            if e.error != const.TELLSTICK_ERROR_DEVICE_NOT_FOUND:
                raise
        return sensors

    def controllers(self):
        """Return all known controllers.

        Requires Telldus core library version >= 2.1.2.

        :return: list of :class:`Controller` instances.
        """
        controllers = []
        try:
            while True:
                controller = self.lib.tdController()
                del controller["name"]
                del controller["available"]
                controllers.append(Controller(lib=self.lib, **controller))
        except TelldusError as e:
            if e.error != const.TELLSTICK_ERROR_NOT_FOUND:
                raise
        return controllers

    def add_device(self, name, protocol, model=None, **parameters):
        """Add a new device.

        :return: a :class:`Device` or :class:`DeviceGroup` instance.
        """
        device = Device(self.lib.tdAddDevice(), lib=self.lib)
        try:
            device.name = name
            device.protocol = protocol
            if model:
                device.model = model
            for key, value in parameters.items():
                device.set_parameter(key, value)

            # Return correct type
            return DeviceFactory(device.id, lib=self.lib)
        except Exception:
            import sys
            exc_info = sys.exc_info()
            try:
                device.remove()
            except:
                pass

            if "with_traceback" in dir(Exception):
                raise exc_info[0].with_traceback(exc_info[1], exc_info[2])
            else:
                exec("raise exc_info[0], exc_info[1], exc_info[2]")

    def add_group(self, name, devices):
        """Add a new device group.

        :return: a :class:`DeviceGroup` instance.
        """
        device = self.add_device(name, "group")
        device.add_to_group(devices)
        return device

    def send_raw_command(self, command, reserved=0):
        """Send a raw command."""
        return self.lib.tdSendRawCommand(command, reserved)

    def connect_controller(self, vid, pid, serial):
        """Connect a controller."""
        self.lib.tdConnectTellStickController(vid, pid, serial)

    def disconnect_controller(self, vid, pid, serial):
        """Disconnect a controller."""
        self.lib.tdDisconnectTellStickController(vid, pid, serial)


def DeviceFactory(id, lib=None):
    """Create the correct device instance based on device type and return it.

    :return: a :class:`Device` or :class:`DeviceGroup` instance.
    """
    lib = lib or Library()
    if lib.tdGetDeviceType(id) == const.TELLSTICK_TYPE_GROUP:
        return DeviceGroup(id, lib=lib)
    return Device(id, lib=lib)


class Device(object):
    """A device that can be controlled by Telldus Core.

    Can be instantiated directly if the id is known, but using
    :func:`DeviceFactory` is recommended. Otherwise returned from
    :func:`TelldusCore.add_device` or :func:`TelldusCore.devices`.
    """

    PARAMETERS = ["devices", "house", "unit", "code", "system", "units",
                  "fade"]

    def __init__(self, id, lib=None):
        super(Device, self).__init__()

        lib = lib or Library()
        super(Device, self).__setattr__('id', id)
        super(Device, self).__setattr__('lib', lib)

    def remove(self):
        """Remove the device from Telldus Core."""
        return self.lib.tdRemoveDevice(self.id)

    def __getattr__(self, name):
        if name == 'name':
            func = self.lib.tdGetName
        elif name == 'protocol':
            func = self.lib.tdGetProtocol
        elif name == 'model':
            func = self.lib.tdGetModel
        elif name == 'type':
            func = self.lib.tdGetDeviceType
        else:
            raise AttributeError(name)
        return func(self.id)

    def __setattr__(self, name, value):
        if name == 'name':
            func = self.lib.tdSetName
        elif name == 'protocol':
            func = self.lib.tdSetProtocol
        elif name == 'model':
            func = self.lib.tdSetModel
        else:
            raise AttributeError(name)
        func(self.id, value)

    def parameters(self):
        """Get dict with all set parameters."""
        parameters = {}
        for name in self.PARAMETERS:
            try:
                parameters[name] = self.get_parameter(name)
            except AttributeError:
                pass
        return parameters

    def get_parameter(self, name):
        """Get a parameter."""
        default_value = "$%!)(INVALID)(!%$"
        value = self.lib.tdGetDeviceParameter(self.id, name, default_value)
        if value == default_value:
            raise AttributeError(name)
        return value

    def set_parameter(self, name, value):
        """Set a parameter."""
        self.lib.tdSetDeviceParameter(self.id, name, str(value))

    def turn_on(self):
        """Turn on the device."""
        self.lib.tdTurnOn(self.id)

    def turn_off(self):
        """Turn off the device."""
        self.lib.tdTurnOff(self.id)

    def bell(self):
        """Send "bell" command to the device."""
        self.lib.tdBell(self.id)

    def dim(self, level):
        """Dim the device.

        :param int level: The level to dim to in the range [0, 255].
        """
        self.lib.tdDim(self.id, level)

    def execute(self):
        """Send "execute" command to the device."""
        self.lib.tdExecute(self.id)

    def up(self):
        """Send "up" command to the device."""
        self.lib.tdUp(self.id)

    def down(self):
        """Send "down" command to the device."""
        self.lib.tdDown(self.id)

    def stop(self):
        """Send "stop" command to the device."""
        self.lib.tdStop(self.id)

    def learn(self):
        """Send "learn" command to the device."""
        self.lib.tdLearn(self.id)

    def methods(self, methods_supported):
        """Query the device for supported methods."""
        return self.lib.tdMethods(self.id, methods_supported)

    def last_sent_command(self, methods_supported):
        """Get the last sent (or seen) command."""
        return self.lib.tdLastSentCommand(self.id, methods_supported)

    def last_sent_value(self):
        """Get the last sent (or seen) value."""
        try:
            return int(self.lib.tdLastSentValue(self.id))
        except ValueError:
            return None


class DeviceGroup(Device):
    """Extends :class:`Device` with methods for managing a group

    E.g. when a group is turned on, all devices in that group are turned on.
    """

    def add_to_group(self, devices):
        """Add device(s) to the group."""
        ids = {d.id for d in self.devices_in_group()}
        ids.update(self._device_ids(devices))
        self._set_group(ids)

    def remove_from_group(self, devices):
        """Remove device(s) from the group."""
        ids = {d.id for d in self.devices_in_group()}
        ids.difference_update(self._device_ids(devices))
        self._set_group(ids)

    def devices_in_group(self):
        """Fetch list of devices in group."""
        try:
            devices = self.get_parameter('devices')
        except AttributeError:
            return []

        ctor = DeviceFactory
        return [ctor(int(x), lib=self.lib) for x in devices.split(',') if x]

    @staticmethod
    def _device_ids(devices):
        try:
            iter(devices)
        except TypeError:
            devices = [devices]

        ids = set()
        for device in devices:
            try:
                ids.add(device.id)
            except AttributeError:
                # Assume device is id
                ids.add(int(device))
        return ids

    def _set_group(self, ids):
        self.set_parameter('devices', ','.join([str(x) for x in ids]))


class Sensor(object):
    """Represents a sensor.

    Returned from :func:`TelldusCore.sensors`
    """

    DATATYPES = {"temperature": const.TELLSTICK_TEMPERATURE,
                 "humidity": const.TELLSTICK_HUMIDITY,
                 "rainrate": const.TELLSTICK_RAINRATE,
                 "raintotal": const.TELLSTICK_RAINTOTAL,
                 "winddirection": const.TELLSTICK_WINDDIRECTION,
                 "windaverage": const.TELLSTICK_WINDAVERAGE,
                 "windgust": const.TELLSTICK_WINDGUST}

    def __init__(self, protocol, model, id, datatypes, lib=None):
        super(Sensor, self).__init__()
        self.protocol = protocol
        self.model = model
        self.id = id
        self.datatypes = datatypes
        self.lib = lib or Library()

    def has_value(self, datatype):
        """Return True if the sensor supports the given data type.

        sensor.has_value(TELLSTICK_TEMPERATURE) is identical to calling
        sensor.has_temperature().
        """
        return (self.datatypes & datatype) != 0

    def value(self, datatype):
        """Return the :class:`SensorValue` for the given data type.

        sensor.value(TELLSTICK_TEMPERATURE) is identical to calling
        sensor.temperature().
        """
        value = self.lib.tdSensorValue(
            self.protocol, self.model, self.id, datatype)
        return SensorValue(datatype, value['value'], value['timestamp'])

    def __getattr__(self, name):
        typename = name.replace("has_", "", 1)
        if typename in Sensor.DATATYPES:
            datatype = Sensor.DATATYPES[typename]
            if name == typename:
                return lambda: self.value(datatype)
            else:
                return lambda: self.has_value(datatype)
        raise AttributeError(name)


class SensorValue(object):
    """Represents a single sensor value.

    Returned from :func:`Sensor.value`.
    """

    __slots__ = ["datatype", "value", "timestamp"]

    def __init__(self, datatype, value, timestamp):
        super(SensorValue, self).__init__()
        self.datatype = datatype
        self.value = value
        self.timestamp = timestamp


class Controller(object):
    """Represents a Telldus controller.

    Returned from :func:`TelldusCore.controllers`
    """

    def __init__(self, id, type, lib=None):
        lib = lib or Library()

        super(Controller, self).__init__()
        super(Controller, self).__setattr__('id', id)
        super(Controller, self).__setattr__('type', type)
        super(Controller, self).__setattr__('lib', lib)

    def __getattr__(self, name):
        try:
            return self.lib.tdControllerValue(self.id, name)
        except TelldusError as e:
            if e.error == const.TELLSTICK_ERROR_METHOD_NOT_SUPPORTED:
                raise AttributeError(name)
            raise

    def __setattr__(self, name, value):
        try:
            self.lib.tdSetControllerValue(self.id, name, value)
        except TelldusError as e:
            if e.error == const.TELLSTICK_ERROR_SYNTAX:
                raise AttributeError(name)
            raise



# Copyright (c) 2012-2014 Erik Johansson <erik@ejohansson.se>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

import unittest

from tellcore.telldus import TelldusCore, TelldusError, Device, DeviceGroup, \
    QueuedCallbackDispatcher
from tellcore.constants import *
import tellcore.library

from ctypes import c_char_p, c_int
import gc
import mocklib


class Test(unittest.TestCase):
    def setUp(self):
        self.mocklib = mocklib.MockTelldusCoreLib()
        self.mockdispatcher = mocklib.MockEventDispatcher()
        self.mockdispatcher.setup_lib_functions(self.mocklib)

        self.loader = mocklib.MockLibLoader(self.mocklib)
        tellcore.library.DllLoader = self.loader

    def tearDown(self):
        self.mockdispatcher.stop()
        del self.mockdispatcher

        tellcore.library.DllLoader = None
        del self.loader
        del self.mocklib

        gc.collect()

        # If a test case fails and throws an exception, the TelldusCore
        # instance will remain in the frame included in the exception which
        # will stop the library from being unloaded. This means that the next
        # test case will reuse the same library instance which causes it to
        # fail.
        if tellcore.library.Library._lib:
            tellcore.library.Library._lib = None
            tellcore.library.Library._refcount = 0

    def decode_string(self, string):
        return string.decode(tellcore.library.Library.STRING_ENCODING)

    def encode_string(self, string):
        return string.encode(tellcore.library.Library.STRING_ENCODING)

    def event_tester(self, core, registrator, trigger, trigger_args):
        event_args = {}

        def callback(*args):
            # Strip away callback id
            event_args[args[-1]] = args[:-1]

        id1 = registrator(callback)
        id2 = registrator(callback)
        id3 = registrator(callback)
        id4 = registrator(callback)
        core.unregister_callback(id4)
        core.unregister_callback(id2)

        trigger(*trigger_args)
        core.callback_dispatcher.process_pending_callbacks()

        callback_args = []
        for arg in trigger_args:
            if type(arg.value) is bytes:
                callback_args.append(self.decode_string(arg.value))
            else:
                callback_args.append(arg.value)
        callback_args = tuple(callback_args)

        self.assertEqual(event_args, {id1: callback_args, id3: callback_args})

    def test_device_event(self):
        core = TelldusCore(callback_dispatcher=QueuedCallbackDispatcher())
        self.event_tester(core, core.register_device_event,
                          self.mockdispatcher.trigger_device_event,
                          (c_int(1), c_int(2), c_char_p(b"foo")))

    def test_device_change_event(self):
        core = TelldusCore(callback_dispatcher=QueuedCallbackDispatcher())
        self.event_tester(core, core.register_device_change_event,
                          self.mockdispatcher.trigger_device_change_event,
                          (c_int(3), c_int(4), c_int(5)))

    def test_raw_device_event(self):
        core = TelldusCore(callback_dispatcher=QueuedCallbackDispatcher())
        self.event_tester(core, core.register_raw_device_event,
                          self.mockdispatcher.trigger_raw_device_event,
                          (c_char_p(b"bar"), c_int(6)))

    def test_sensor_event(self):
        core = TelldusCore(callback_dispatcher=QueuedCallbackDispatcher())
        self.event_tester(core, core.register_sensor_event,
                          self.mockdispatcher.trigger_sensor_event,
                          (c_char_p(b"proto"), c_char_p(b"model"), c_int(7),
                           c_int(8), c_char_p(b"value"), c_int(9)))

    def test_controller_event(self):
        core = TelldusCore(callback_dispatcher=QueuedCallbackDispatcher())
        self.event_tester(core, core.register_controller_event,
                          self.mockdispatcher.trigger_controller_event,
                          (c_int(10), c_int(11), c_int(12), c_char_p(b"new")))

    def test_group(self):
        self.mocklib.tdAddDevice = lambda: 1
        self.mocklib.tdGetDeviceType = lambda id: TELLSTICK_TYPE_GROUP

        device = {}
        device['parameters'] = {}

        def tdSetName(id, name):
            device['name'] = self.decode_string(name)
            return id == 1
        self.mocklib.tdSetName = tdSetName

        def tdSetProtocol(id, protocol):
            device['protocol'] = self.decode_string(protocol)
            return id == 1
        self.mocklib.tdSetProtocol = tdSetProtocol

        def tdSetDeviceParameter(id, name, value):
            name = self.decode_string(name)
            if value:
                device['parameters'][name] = self.decode_string(value)
            elif name in device['parameters']:
                del device['parameters'][name]
            return id == 1
        self.mocklib.tdSetDeviceParameter = tdSetDeviceParameter

        def tdGetDeviceParameter(id, name, default):
            name = self.decode_string(name)
            try:
                return c_char_p(self.encode_string(device['parameters'][name]))
            except KeyError:
                return c_char_p(default)
        self.mocklib.tdGetDeviceParameter = tdGetDeviceParameter

        core = TelldusCore()
        dev = core.add_group("test", [1, Device(2), 3, 4])

        self.assertEqual("test", device['name'])
        self.assertEqual("group", device['protocol'])
        self.assertDictEqual({'devices': '1,2,3,4'}, device['parameters'])

        dev.add_to_group(5)
        self.assertDictEqual({'devices': '1,2,3,4,5'}, device['parameters'])

        dev.remove_from_group([3, 2, 6, Device(2), Device(1)])
        self.assertDictEqual({'devices': '4,5'}, device['parameters'])
        self.assertListEqual([4, 5], [d.id for d in dev.devices_in_group()])

        dev.remove_from_group(4)
        self.assertDictEqual({'devices': '5'}, device['parameters'])
        self.assertIsInstance(dev.devices_in_group()[0], DeviceGroup)

        dev.remove_from_group([5])
        self.assertDictEqual({}, device['parameters'])

    def test_devices(self):
        devs = {0: {'protocol': b"proto_1", 'model': b"model_1"},
                3: {'protocol': b"proto_2", 'model': b"model_2"},
                6: {'protocol': b"proto_3", 'model': b"model_3"}}

        self.mocklib.tdGetNumberOfDevices = lambda: len(devs)
        self.mocklib.tdGetDeviceId = lambda index: index * 3
        self.mocklib.tdGetDeviceType = lambda id: TELLSTICK_TYPE_DEVICE
        self.mocklib.tdGetProtocol = lambda id: \
            c_char_p(devs[id]['protocol'])
        self.mocklib.tdGetModel = lambda id: \
            c_char_p(devs[id]['model'])

        core = TelldusCore()
        devices = core.devices()

        self.assertEqual(3, len(devices))
        self.assertEqual(['proto_1', 'proto_2', 'proto_3'],
                         [d.protocol for d in devices])
        self.assertEqual(['model_1', 'model_2', 'model_3'],
                         [d.model for d in devices])

    def test_device(self):
        def actor(id):
            if id == 3:
                return TELLSTICK_SUCCESS
            else:
                return TELLSTICK_ERROR_DEVICE_NOT_FOUND
        self.mocklib.tdTurnOn = actor
        self.mocklib.tdTurnOff = actor
        self.mocklib.tdBell = lambda id: TELLSTICK_ERROR_METHOD_NOT_SUPPORTED

        device = Device(3)
        device.turn_on()
        device.turn_off()
        self.assertRaises(TelldusError, device.bell)

        device = Device(4)
        self.assertRaises(TelldusError, device.turn_on)
        self.assertRaises(TelldusError, device.turn_off)
        self.assertRaises(TelldusError, device.bell)

    def test_sensors(self):
        self.sensor_index = 0

        def tdSensor(protocol, p_len, model, m_len, id, datatypes):
            sensors = [{'protocol': b"proto_1", 'model': b"model_1", 'id': 1,
                        'datatypes': TELLSTICK_TEMPERATURE},
                       {'protocol': b"proto_2", 'model': b"model_2", 'id': 2,
                        'datatypes': TELLSTICK_HUMIDITY},
                       {'protocol': b"proto_3", 'model': b"model_3", 'id': 3,
                        'datatypes': TELLSTICK_RAINRATE},
                       {'protocol': b"proto_4", 'model': b"model_4", 'id': 4,
                        'datatypes': TELLSTICK_RAINTOTAL},
                       {'protocol': b"proto_5", 'model': b"model_5", 'id': 5,
                        'datatypes': TELLSTICK_WINDDIRECTION},
                       {'protocol': b"proto_6", 'model': b"model_6", 'id': 6,
                        'datatypes': TELLSTICK_WINDAVERAGE},
                       {'protocol': b"proto_7", 'model': b"model_7", 'id': 7,
                        'datatypes': TELLSTICK_WINDGUST}]
            if self.sensor_index < len(sensors):
                sensor = sensors[self.sensor_index]
                self.sensor_index += 1

                protocol.value = sensor['protocol']
                model.value = sensor['model']
                try:
                    id._obj.value = sensor['id']
                    datatypes._obj.value = sensor['datatypes']
                except AttributeError:
                    # With pypy we must use contents instead of _obj
                    id.contents.value = sensor['id']
                    datatypes.contents.value = sensor['datatypes']
                return TELLSTICK_SUCCESS
            else:
                self.sensor_index = 0
                return TELLSTICK_ERROR_DEVICE_NOT_FOUND
        self.mocklib.tdSensor = tdSensor

        def tdSensorValue(protocol, model, id, datatype, value, v_len, times):
            if datatype == 1 << (id - 1):
                value.value = self.encode_string("%d" % (id * 100 + datatype))
                return TELLSTICK_SUCCESS
            else:
                return TELLSTICK_ERROR_METHOD_NOT_SUPPORTED
        self.mocklib.tdSensorValue = tdSensorValue

        core = TelldusCore()
        sensors = core.sensors()

        self.assertEqual(7, len(sensors))
        self.assertEqual(["proto_%d" % i for i in range(1, 8)],
                         [s.protocol for s in sensors])
        self.assertEqual(["model_%d" % i for i in range(1, 8)],
                         [s.model for s in sensors])
        self.assertEqual(list(range(1, 8)),
                         [s.id for s in sensors])
        self.assertEqual([TELLSTICK_TEMPERATURE, TELLSTICK_HUMIDITY,
                          TELLSTICK_RAINRATE, TELLSTICK_RAINTOTAL,
                          TELLSTICK_WINDDIRECTION, TELLSTICK_WINDAVERAGE,
                          TELLSTICK_WINDGUST],
                         [s.datatypes for s in sensors])

        self.assertEqual([False] * 0 + [True] + [False] * 6,
                         [s.has_temperature() for s in sensors])
        self.assertEqual([False] * 1 + [True] + [False] * 5,
                         [s.has_humidity() for s in sensors])
        self.assertEqual([False] * 2 + [True] + [False] * 4,
                         [s.has_rainrate() for s in sensors])
        self.assertEqual([False] * 3 + [True] + [False] * 3,
                         [s.has_raintotal() for s in sensors])
        self.assertEqual([False] * 4 + [True] + [False] * 2,
                         [s.has_winddirection() for s in sensors])
        self.assertEqual([False] * 5 + [True] + [False] * 1,
                         [s.has_windaverage() for s in sensors])
        self.assertEqual([False] * 6 + [True] + [False] * 0,
                         [s.has_windgust() for s in sensors])

        self.assertEqual("%d" % (100 + TELLSTICK_TEMPERATURE),
                         sensors[0].temperature().value)
        self.assertEqual("%d" % (200 + TELLSTICK_HUMIDITY),
                         sensors[1].humidity().value)
        self.assertEqual("%d" % (300 + TELLSTICK_RAINRATE),
                         sensors[2].rainrate().value)
        self.assertEqual("%d" % (400 + TELLSTICK_RAINTOTAL),
                         sensors[3].raintotal().value)
        self.assertEqual("%d" % (500 + TELLSTICK_WINDDIRECTION),
                         sensors[4].winddirection().value)
        self.assertEqual("%d" % (600 + TELLSTICK_WINDAVERAGE),
                         sensors[5].windaverage().value)
        self.assertEqual("%d" % (700 + TELLSTICK_WINDGUST),
                         sensors[6].windgust().value)

if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2012-2014 Erik Johansson <erik@ejohansson.se>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

import unittest

import tellcore.library
from tellcore.constants import *

import ctypes
import gc
import mocklib

Library = tellcore.library.Library
TelldusError = tellcore.library.TelldusError


class Test(unittest.TestCase):
    def setUp(self):
        self.initialized = False
        self.mocklib = mocklib.MockTelldusCoreLib()

        def tdInit():
            if self.initialized:
                raise RuntimeError("already initialized")
            self.initialized = True
        self.mocklib.tdInit = tdInit

        def tdClose():
            if not self.initialized:
                raise RuntimeError("not initialized")
            self.initialized = False
        self.mocklib.tdClose = tdClose

        self.loader = mocklib.MockLibLoader(self.mocklib)
        tellcore.library.DllLoader = self.loader

    def tearDown(self):
        gc.collect()

    def test_libloader(self):
        self.assertEqual(self.loader.load_count, 0)
        lib = Library()
        self.assertEqual(self.loader.load_count, 1)
        self.assertEqual(lib._lib, self.mocklib)

    def test_oneinstance(self):
        """Test that the singleton works"""
        lib1 = Library()
        lib2 = Library()
        self.assertEqual(self.loader.load_count, 1)

        del lib1, lib2
        gc.collect()

        lib = Library()
        self.assertEqual(self.loader.load_count, 2)
        self.assertIsInstance(lib, Library)

    def test_initialized(self):
        self.assertFalse(self.initialized)
        lib1 = Library()
        self.assertTrue(self.initialized)
        lib2 = Library()
        self.assertTrue(self.initialized)

        del lib1
        gc.collect()
        self.assertTrue(self.initialized)

        del lib2
        gc.collect()
        self.assertFalse(self.initialized)

    def test_private_methods(self):
        lib = Library()
        self.assertRaises(NotImplementedError, lib.tdInit)
        self.assertRaises(NotImplementedError, lib.tdClose)
        self.assertRaises(NotImplementedError, lib.tdReleaseString, 0)

    def test_string_releaser(self):
        """Test that all strings returned from core lib are released"""
        released = []

        def tdReleaseString(pointer):
            released.append(pointer.value)
        self.mocklib.tdReleaseString = tdReleaseString

        returned = []

        def tdGetErrorString(error):
            string = ctypes.c_char_p(
                ("error %d" % error).encode(Library.STRING_ENCODING))
            void_pointer = ctypes.cast(string, ctypes.c_void_p)
            returned.append(void_pointer.value)
            return string
        self.mocklib.tdGetErrorString = tdGetErrorString

        lib = Library()
        for i in range(-5, 0):
            lib.tdGetErrorString(i)

        self.assertEqual(len(released), 5)
        self.assertEqual(returned, released)

    def test_string_parameter(self):
        def int_strings(ret, ignore, string1, string2="defaultValue"):
            self.assertEqual(string1, b"aString")
            if string2 != "defaultValue":
                self.assertEqual(string2, b"aSecondString")
            return ret

        string_int = lambda s, i: int_strings(0, i, s)
        ints_string = lambda i1, i2, s: int_strings(0, i1, s)
        strings_ignore = lambda s1, s2, *args: int_strings(0, 0, s1, s2)
        int_string_ignore = lambda i, s, *args: int_strings(0, i, s)

        self.mocklib.tdSetName = lambda *a: int_strings(True, *a)
        self.mocklib.tdSetProtocol = lambda *a: int_strings(True, *a)
        self.mocklib.tdSetModel = lambda *a: int_strings(True, *a)
        self.mocklib.tdSetDeviceParameter = lambda *a: int_strings(True, *a)
        self.mocklib.tdGetDeviceParameter = lambda *a: int_strings(0, *a)
        self.mocklib.tdSendRawCommand = string_int
        self.mocklib.tdConnectTellStickController = ints_string
        self.mocklib.tdDisconnectTellStickController = ints_string
        self.mocklib.tdSensorValue = strings_ignore
        self.mocklib.tdControllerValue = int_string_ignore
        self.mocklib.tdSetControllerValue = lambda *a: int_strings(0, *a)

        lib = Library()

        lib.tdSetName(1, "aString")
        lib.tdSetProtocol(1, "aString")
        lib.tdSetModel(1, "aString")
        lib.tdSetDeviceParameter(1, "aString", "aSecondString")
        lib.tdGetDeviceParameter(1, "aString", "aSecondString")
        lib.tdSendRawCommand("aString", 1)
        lib.tdConnectTellStickController(0, 0, "aString")
        lib.tdDisconnectTellStickController(0, 0, "aString")
        lib.tdSensorValue("aString", "aSecondString", 0, 0)
        lib.tdControllerValue(1, "aString")
        lib.tdSetControllerValue(1, "aString", "aSecondString")

    def test_unicode(self):
        test_name = b'\xc3\xa5\xc3\xa4\xc3\xb6'

        def tdSetName(id, name):
            self.assertIs(type(name), bytes)
            self.assertEqual(name, test_name)
            return True
        self.mocklib.tdSetName = tdSetName

        lib = Library()

        lib.tdSetName(1, test_name.decode(Library.STRING_ENCODING))

    def setup_callback(self, registered_ids, unregistered_ids):
        def tdRegisterEvent(*args):
            cid = len(registered_ids) + 1
            registered_ids.append(cid)
            return cid
        self.mocklib.tdRegisterDeviceEvent = tdRegisterEvent
        self.mocklib.tdRegisterDeviceChangeEvent = tdRegisterEvent
        self.mocklib.tdRegisterRawDeviceEvent = tdRegisterEvent
        self.mocklib.tdRegisterSensorEvent = tdRegisterEvent
        self.mocklib.tdRegisterControllerEvent = tdRegisterEvent

        def tdUnregisterCallback(cid):
            unregistered_ids.append(cid)
            return TELLSTICK_SUCCESS
        self.mocklib.tdUnregisterCallback = tdUnregisterCallback

    def test_callback_cleanup(self):
        registered_ids = []
        unregistered_ids = []
        self.setup_callback(registered_ids, unregistered_ids)

        def callback(*args):
            pass

        dispatcher = tellcore.library.DirectCallbackDispatcher
        lib = Library(callback_dispatcher=dispatcher)
        lib.tdRegisterDeviceEvent(callback)
        lib.tdRegisterDeviceChangeEvent(callback)
        lib.tdRegisterRawDeviceEvent(callback)
        lib.tdRegisterSensorEvent(callback)
        lib.tdRegisterControllerEvent(callback)

        self.assertEqual(len(registered_ids), 5)
        self.assertEqual(len(unregistered_ids), 0)

        lib = None
        gc.collect()
        self.assertEqual(registered_ids, unregistered_ids)

    def test_callback_not_shared(self):
        registered_ids = []
        unregistered_ids = []
        self.setup_callback(registered_ids, unregistered_ids)

        def callback(*args):
            pass

        dispatcher = tellcore.library.DirectCallbackDispatcher
        lib = Library(callback_dispatcher=dispatcher)

        cid = lib.tdRegisterDeviceEvent(callback)
        self.assertEqual(registered_ids, [cid])
        self.assertEqual(unregistered_ids, [])

        lib_copy = Library(callback_dispatcher=dispatcher)

        cid_copy = lib_copy.tdRegisterDeviceEvent(callback)
        self.assertEqual(registered_ids, [cid, cid_copy])
        self.assertEqual(unregistered_ids, [])

        lib = None
        gc.collect()
        self.assertEqual(unregistered_ids, [cid])

        lib_copy = None
        gc.collect()
        self.assertEqual(registered_ids, unregistered_ids)

    def test_exception_on_error(self):
        def tdGetNumberOfDevices():
            return TELLSTICK_ERROR_CONNECTING_SERVICE
        self.mocklib.tdGetNumberOfDevices = tdGetNumberOfDevices

        lib = Library()
        with self.assertRaises(TelldusError) as cm:
            lib.tdGetNumberOfDevices()
        self.assertEqual(cm.exception.error,
                         TELLSTICK_ERROR_CONNECTING_SERVICE)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

from distutils.core import setup
import os
import re

version_re = re.compile(r'__version__ = "(.*)"')
cwd = os.path.dirname(os.path.abspath(__file__))

with open(os.path.join(cwd, 'tellcore', '__init__.py')) as init:
    for line in init:
        match = version_re.search(line)
        if match:
            version = match.group(1)
            break
    else:
        raise Exception('Cannot find version in __init__.py')

setup(
    name='tellcore-py',
    version=version,
    author='Erik Johansson',
    author_email='erik@ejohansson.se',
    packages=['tellcore'],
    provides=['tellcore'],
    scripts=['bin/tellcore_tool', 'bin/tellcore_events',
             'bin/tellcore_controllers'],
    url='https://github.com/erijo/tellcore-py',
    license='GPLv3+',
    description='Python wrapper for Telldus\' home automation library',
    long_description=open('README.rst').read() + '\n\n' + \
        open('CHANGES.rst').read(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Home Automation',
        'Topic :: Software Development :: Libraries :: Python Modules',
        ],
)



# Copyright (c) 2012-2014 Erik Johansson <erik@ejohansson.se>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

try:
    import queue
except ImportError:
    import Queue as queue

import ctypes
import threading

import tellcore.library

ByRefArgType = type(ctypes.byref(ctypes.c_int(0)))


class MockLibLoader(object):
    def __init__(self, mocklib):
        object.__init__(self)
        self.load_count = 0
        self.mocklib = mocklib

    def LoadLibrary(self, name):
        self.load_count += 1
        return self.mocklib


class MockTelldusCoreLib(object):
    def __init__(self):
        object.__init__(self)

        self.tdInit = lambda: None
        self.tdClose = lambda: None

        self.tdReleaseString = lambda x: None
        self.tdGetErrorString = lambda x: None

    def __getattr__(self, name):
        if name in tellcore.library.Library._functions:
            func = MockCFunction(name, self)
            object.__setattr__(self, name, func)
            return func
        raise AttributeError(name)

    def __setattr__(self, name, value):
        if name in tellcore.library.Library._functions:
            func = getattr(self, name)
            func.implementation = value
        else:
            object.__setattr__(self, name, value)


class MockCFunction(object):
    def __init__(self, name, lib):
        object.__init__(self)
        self.name = name
        self.lib = lib
        self.implementation = None
        self.restype = None
        self.argtypes = None
        self.errcheck = None

    def __call__(self, *args):
        if self.implementation is None:
            raise NotImplementedError("%s is not implemented" % self.name)

        if self.argtypes is None:
            raise NotImplementedError("%s not configured" % self.name)

        if len(self.argtypes) != len(args):
            raise TypeError("%s() takes exactly %d argument(s) (%d given)" %
                            (self.name, len(self.argument), len(args)))

        c_args = []

        # Verify that the arguments are of correct type
        for c_type, value in zip(self.argtypes, args):
            c_args.append(value)
            if type(value) is not c_type:
                # The 'raw' attribute is the pointer for string buffers
                if hasattr(value, 'raw'):
                    c_type.from_param(value.raw)
                elif type(value) is not ByRefArgType:
                    c_type.from_param(value)
                    # Pass the possibly converted value instead of the original
                    c_args[-1] = c_type(value).value

        res = self.implementation(*c_args)

        # Functions returning char pointers are set up to return the pointer as
        # a void pointer. Do the conversion here to match the real thing.
        if self.restype is ctypes.c_void_p:
            res = ctypes.cast(res, ctypes.c_void_p)
        # For the rest, verify that the return value is of correct type.
        elif self.restype is not None:
            self.restype.from_param(res)

        if self.errcheck is not None:
            res = self.errcheck(res, self, args)
        return res


class MockEventDispatcher(threading.Thread):
    def __init__(self):
        threading.Thread.__init__(self, name="MockEventDispatcher")

        self.next_callback_id = 1

        # callback id -> callback object
        self.callbacks = {}

        # Lists of callback ids
        self.device_callbacks = []
        self.device_change_callbacks = []
        self.raw_device_callbacks = []
        self.sensor_callbacks = []
        self.controller_callbacks = []

        self.queue = queue.Queue()
        self.running = True
        self.start()

    def setup_lib_functions(self, mocklib):
        def tdUnregisterCallback(cid):
            del self.callbacks[cid]
            return 1
        mocklib.tdUnregisterCallback = tdUnregisterCallback

        def register_callback(active_list, callback, context):
            cid = self.next_callback_id
            self.next_callback_id += 1

            self.callbacks[cid] = (callback, context)
            active_list.append(cid)
            return cid

        callback_types = {
            "tdRegisterDeviceEvent": self.device_callbacks,
            "tdRegisterDeviceChangeEvent": self.device_change_callbacks,
            "tdRegisterRawDeviceEvent": self.raw_device_callbacks,
            "tdRegisterSensorEvent": self.sensor_callbacks,
            "tdRegisterControllerEvent": self.controller_callbacks
        }

        for name, id_list in callback_types.items():
            func = lambda callback, context, id_list=id_list: \
                register_callback(id_list, callback, context)
            setattr(mocklib, name, func)

    def run(self):
        while self.running:
            (callback, args) = self.queue.get()
            try:
                callback(*args)
            except:
                pass
            finally:
                self.queue.task_done()

    def stop(self):
        assert self.running

        def stop_thread():
            self.running = False
        self.queue_event(stop_thread)
        self.queue.join()

    def queue_event(self, callback, *args):
        assert self.running
        self.queue.put((callback, args))

    def _trigger_event(self, callback_list, *args):
        for cid in callback_list:
            if cid in self.callbacks:
                (callback, context) = self.callbacks[cid]
                event_args = args + (cid, context)
                self.queue_event(callback, *event_args)
        # Make sure all events are delivered
        self.queue.join()

    def trigger_device_event(self, device_id, method, data):
        self._trigger_event(self.device_callbacks, device_id, method, data)

    def trigger_device_change_event(self, device_id, event, type_):
        self._trigger_event(self.device_change_callbacks, device_id,
                            event, type_)

    def trigger_raw_device_event(self, data, controller_id):
        self._trigger_event(self.raw_device_callbacks, data, controller_id)

    def trigger_sensor_event(self, protocol, model, sensor_id, dataType, value,
                             timestamp):
        self._trigger_event(self.sensor_callbacks, protocol, model, sensor_id,
                            dataType, value, timestamp)

    def trigger_controller_event(self, controller_id, event, type_, new_value):
        self._trigger_event(self.controller_callbacks, controller_id, event,
                            type_, new_value)
